/*
 * PDBExporter.java
 * Mark Cushman (mark@cushman.net)
 * Created on January 16, 2003, 4:47 PM
 */

package staging;

import java.io.*;
import java.util.Date;
import java.util.Vector;
import java.util.Calendar;
import java.util.StringTokenizer;

public class PDBExporter {
    
    //constants
    private String  PDB_NAME            = "Palantir Music Database";
    private String  PDB_TYPE            = "DATA";
    private String  PDB_CREATOR         = "EmPC";
    private int     HEADER_SIZE         = 78;
    
    //vectors of records
    private Vector  artistRecords;
    private Vector  trackRecords;
    
    //pdb header structure
    private byte    name[];             //set in writeHeader
    private byte    filler[]            = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    private short   attributes          = 0;
    private short   version             = 1;
    private int     creationDate;       //set in writeHeader
    private int     modificationDate;   //set in writeHeader
    private int     lastBackupDate;     //set in writeHeader
    private int     modificationNumber  = 1;
    private int     appInfoID           = 0;
    private int     sortInfoID          = 0;
    private byte    type[];             //set in writeHeader
    private byte    creator[];          //set in writeHeader
    private int     uniqueIDSeed        = 0;
    private int     nextRecord          = 0;
    
    public void write(File file, OutputStream _os) throws IOException {
        //this is what we need write to look like later...
        //public void write(PlayerDatabase _db, OutputStream _os) throws IOException
        
        //set up our output file datastream
        DataOutputStream os = new DataOutputStream(_os);
        
        //populate our records Vector with lines from the csv file
        System.out.println("Creating records from .csv file...");
        populateRecords(file);
        
        //write the header out to the .pdb file
        writeHeader(os);
        
        //write the record list structure to the .pdb file
        writeRecordList(os);
        
        //now write out the records
        //write the header first
        os.writeInt(1);                         //header: artistRecordStart
        os.writeInt(artistRecords.size());      //header: artistRecordLength
        os.writeInt(artistRecords.size() + 1);  //header: trackRecordStart
        os.writeInt(trackRecords.size());       //header: trackRecordLength
        
        //now write the artist records
        for (int i=0; i<artistRecords.size(); i++)
            writeArtistRecord(os, (PDBArtistRecord)artistRecords.get(i));
        
        //lastly come the track records
        for (int i=0; i<trackRecords.size(); i++)
            writeTrackRecord(os, (PDBTrackRecord)trackRecords.get(i));
        
        os.close();
    }
    
    private void populateRecords(File file) throws IOException {
        BufferedReader in = new BufferedReader(new FileReader(file));
        
        PDBArtistRecord artistRecord = null;
        PDBTrackRecord trackRec = null;
        String oldSource = null, oldArtist = null;
        int trackCounter = 0, artistCounter = 0,
        sourceCounter = 0, skipped = 0, currentTracks = 0;
        int[] sourceOffset = new int[1];
        
        artistRecords = new Vector();
        trackRecords = new Vector();
        
        //read each line into our records Vector
        String line = in.readLine();
        while(line != null) {
            
            StringTokenizer st = new StringTokenizer(line, ",");
            if(st.countTokens() == 9) {
                //get the data from the line
                String name = st.nextToken() + "\0";
                String year = st.nextToken() + "\0";
                String source = st.nextToken() + "\0";
                String track = st.nextToken() + "\0";
                String title = st.nextToken() + "\0";
                String genre = st.nextToken() + "\0";
                String bitrate = st.nextToken() + "\0";
                String duration = st.nextToken();

                long lduration = Long.parseLong(duration);
                long seconds = (lduration / 1000) % 60;
                long minutes = (lduration / (60 * 1000)) % 60;
                long hours = (lduration / (60 * 60 * 1000)) % 60;

                Long secs = new Long(seconds);
                Long mins = new Long(minutes);
                Long hrs = new Long(hours);
                
                duration = (hours > 0 ? hrs.toString() + ":" : "");
                if(duration.equals(""))
                    duration += mins.toString() + ":";
                else
                    duration += (minutes < 10 ? "0" + mins.toString() : mins.toString()) + ":";
                duration += (seconds < 10 ? "0" + secs.toString() : secs.toString()) + "\0";
               
                Integer fidInt = new Integer(st.nextToken());
                int fid = fidInt.intValue();
                
                if(!name.equals(oldArtist)) {
                    //we have a new artist
                    if(artistRecord != null) {
                        //first complete the last artist record
                        artistRecord.trackCount = currentTracks;
                        artistRecord.sourceCount = sourceCounter;
                        artistRecord.sourceOffset = sourceOffset;
                        artistRecords.add(artistRecord);
                        System.out.println("\nStored Artist: " + artistRecord.name);
                        System.out.println("id:            " + artistRecord.artistID);
                        System.out.println("trackcount:   " + artistRecord.trackCount);
                        System.out.println("sourcecount:   " + artistRecord.sourceCount);
                        System.out.print("sourceoffsets: ");
                        for(int i=0; i<artistRecord.sourceOffset.length; i++) {
                            System.out.print(artistRecord.sourceOffset[i] + "  ");
                        }
                        System.out.println("");
                    }
                    
                    //now make a new artist record
                    artistRecord = new PDBArtistRecord();
                    artistRecord.name = name;
                    oldArtist = artistRecord.name;
                    oldSource = "";
                    artistRecord.artistID = artistCounter;
                    currentTracks = 0;
                    sourceCounter = 0;
                    sourceOffset = new int[1];
                    artistCounter++;
                }
                
                if(!source.equals(oldSource)) {
                    sourceCounter++;
                    if(sourceCounter > 1) {
                        //grow our array
                        int[] tmpArray = new int[sourceCounter];
                        System.arraycopy(sourceOffset, 0, tmpArray, 0, sourceOffset.length);
                        sourceOffset = new int[sourceCounter];
                        System.arraycopy(tmpArray, 0, sourceOffset, 0, tmpArray.length);
                    }
                    sourceOffset[sourceCounter - 1] = trackCounter;
                    oldSource = source;
                }
                
                trackRec = new PDBTrackRecord();
                trackRec.fid = fid;
                trackRec.artistID = artistRecord.artistID;
                trackRec.data = year + source + track + title + genre + bitrate + duration;
                
                //print out our record
                System.out.println("");         System.out.println("Record " + trackCounter);
                System.out.print("fid:      "); System.out.println(trackRec.fid);
                System.out.print("artistID: "); System.out.println(trackRec.artistID);
                System.out.print("data:     "); System.out.println(trackRec.data);
                
                currentTracks++;
                trackCounter++;
                trackRecords.add(trackRec);
                line = in.readLine();
            } else {
                //a line with commas
                skipped++;
                line = in.readLine();
            }
        }
        
        //finish up the last artist we have to store
        artistRecord.trackCount = currentTracks;
        artistRecord.sourceCount = sourceCounter;
        artistRecord.sourceOffset = sourceOffset;
        artistRecords.add(artistRecord);
        System.out.println("\nStored Artist: " + artistRecord.name);
        System.out.println("id:            " + artistRecord.artistID);
        System.out.println("trackcount:   " + artistRecord.trackCount);
        System.out.println("sourcecount:   " + artistRecord.sourceCount);
        System.out.print("sourceoffsets: ");
        for(int i=0; i<artistRecord.sourceOffset.length; i++) {
            System.out.print(artistRecord.sourceOffset[i] + "  ");
        }
        System.out.println("");
        
        //we done now, yahear?
        in.close();
        System.out.println("\nSkipped " + skipped + " records.");
    }
    
    private void writeHeader(DataOutputStream os) throws IOException {
        
        //get the current date in seconds from Jan 1, 1904
        Calendar cal = Calendar.getInstance();
        cal.set(2036, 0, 1);
        Date date = new Date();
        int now = (int)(date.getTime() + cal.getTime().getTime())/1000;
        
        //now set up all our variables
        name                = PDB_NAME.getBytes();
        type                = PDB_TYPE.getBytes();
        creator             = PDB_CREATOR.getBytes();
        creationDate        = now;
        modificationDate    = now;
        lastBackupDate      = now;
        
        //write the header out to the file
        os.write(name);
        os.write(filler, 0, 32 - name.length);
        os.writeShort(attributes);
        os.writeShort(version);
        os.writeInt(creationDate);
        os.writeInt(modificationDate);
        os.writeInt(lastBackupDate);
        os.writeInt(modificationNumber);
        os.writeInt(appInfoID);
        os.writeInt(sortInfoID);
        os.write(type);
        os.write(creator);
        os.writeInt(uniqueIDSeed);
        
        //write the record list structure
        os.writeInt(nextRecord);
        //header + # of artists + # of tracks
        os.writeShort(1 + artistRecords.size() + trackRecords.size());
    }
    
    private void writeRecordList(DataOutputStream os) throws IOException {
        
        //calculate the offset
        int j = 0;
        int offset = HEADER_SIZE;
        offset += (8 * (1 + artistRecords.size() + trackRecords.size())) + 2;
        
        //first write out our header record
        os.writeInt(offset);
        os.writeByte(0);                // attributes
        os.writeByte((byte)(j >> 16));  // uniqueID
        os.writeByte((byte)(j >> 8));   // uniqueID
        os.writeByte((byte)(j >> 0));   // uniqueID
        //increment the offset size
        offset += 16;  //16 bytes for header
        
        //now write the record list structure for Artists
        for (int i=0; i<artistRecords.size(); i++) {
            j++;
            os.writeInt(offset);
            os.writeByte(0);                // attributes
            os.writeByte((byte)(j >> 16));  // uniqueID
            os.writeByte((byte)(j >> 8));   // uniqueID
            os.writeByte((byte)(j >> 0));   // uniqueID
            //increment the offset size
            PDBArtistRecord tmpRec = (PDBArtistRecord)artistRecords.get(i);
            offset += tmpRec.size();
        }
        
        //now write the record list structure for Tracks
        for (int i=0; i<trackRecords.size(); i++) {
            j++;
            os.writeInt(offset);
            os.writeByte(0);                // attributes
            os.writeByte((byte)(j >> 16));  // uniqueID
            os.writeByte((byte)(j >> 8));   // uniqueID
            os.writeByte((byte)(j >> 0));   // uniqueID
            //increment the offset size
            PDBTrackRecord tmpRec = (PDBTrackRecord)trackRecords.get(i);
            offset += tmpRec.size();
        }
        
        os.writeShort(0);                   // padding
    }
    
    private void writeArtistRecord(DataOutputStream os, PDBArtistRecord artistRec) throws IOException {
        os.writeInt(artistRec.artistID);
        os.writeInt(artistRec.trackCount);
        os.writeInt(artistRec.sourceCount);
        for(int i=0; i<artistRec.sourceOffset.length; i++)
            os.writeInt(artistRec.sourceOffset[i]);
        os.write(artistRec.name.getBytes());
    }
    
    private void writeTrackRecord(DataOutputStream os, PDBTrackRecord trackRec) throws IOException {
        os.writeInt(trackRec.fid);
        os.writeInt(trackRec.artistID);
        os.write(trackRec.data.getBytes());
    }
    
    public static void main(String[] args) {
        PDBExporter pdbex = new PDBExporter();
        try {
            pdbex.write(new File("D:\\projects\\palantir\\data\\test.csv"), new FileOutputStream("D:\\projects\\palantir\\data\\palantirData.pdb"));
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
}